'''
--------------------------------------------------------------------------------------
General psychoacoustic tools for working with music synthesis.

`Aikous` is a specialized module for working with psychoacoustics in the context of music.

see: https://en.wikipedia.org/wiki/Psychoacoustics

The word "aikous" is a portmanteau derived from Ancient Greek, blending the elements of 
"αἰσθάνομαι" (aisthanomai), meaning "to perceive" or "to feel," and "ἀκούω" (akouo), 
meaning "to hear."

The `aikous` module contains tools for translating physics phenomena, as perceived by
humans, into algebraic musical representations.
--------------------------------------------------------------------------------------
'''
from .dynamics import Dynamic, DynamicRange, dbamp, ampdb, freq_amp_scale
from .enevelopes import line, arch, map_curve

__all__ = [
    'Dynamic',
    'DynamicRange',
    'dbamp',
    'ampdb',
    'freq_amp_scale',
    'line',
    'arch',
    'map_curve'
]


import numpy as np
from itertools import combinations
import pandas as pd
import math
from fractions import Fraction
from typing import List, Tuple, Set, Dict, Any, Union
import networkx as nx
import sympy as sp


__all__ = [
    'Operations',
    'Sieve',
    'CombinationSet',
    'PartitionSet',
    'GenCol',
]

# ------------------------------------------------------------------------------
# Set Operations
# --------------

class Operations:
    '''
    Class for set operations.
    '''
    @staticmethod
    def union(set1: set, set2: set) -> set:        
        '''Return the union of two sets.'''
        return set1 | set2

    @staticmethod
    def intersect(set1: set, set2: set) -> set:
        '''Return the intersection of two sets.'''
        return set1 & set2

    @staticmethod
    def diff(set1: set, set2: set) -> set:
        '''Return the difference of two sets (elements in set1 but not in set2).'''
        return set1 - set2

    @staticmethod
    def symm_diff(set1: set, set2: set) -> set:
        '''Return the symmetric difference of two sets (elements in either set1 or set2 but not both).'''
        return set1 ^ set2

    @staticmethod
    def is_subset(subset: set, superset: set) -> bool:
        '''Check if the first set is a subset of the second set.'''
        return subset <= superset

    @staticmethod
    def is_superset(superset: set, subset: set) -> bool:
        '''Check if the first set is a superset of the second set.'''
        return superset >= subset

    @staticmethod
    def invert(set1: set, axis: int = 0, modulus: int = 12) -> set:
        '''Invert a set around a given axis using modular arithmetic.'''
        return {(axis * 2 - pitch) % modulus for pitch in set1}

    @staticmethod
    def transpose(set1: set, transposition_interval: int, modulus: int = 12) -> set:
        '''Transpose a set by a given interval using modular arithmetic.'''
        return {(pitch + transposition_interval) % modulus for pitch in set1}

    @staticmethod
    def complement(S: set, modulus: int = 12) -> set:
        '''Return the complement of a set within a given modulus.'''
        return {s for s in range(modulus) if s not in S}

    @staticmethod
    def congruent(S: set, modulus: int, residue: int) -> set:
        '''Return the set of all values in set1 that are congruent modulo the given modulus and residue.'''
        return {s for s in S if s % modulus == residue}

    @staticmethod
    def intervals(S: set) -> set:
        '''
        Calculate the set of intervals between successive numbers in a sorted sequence.

        Args:
            numbers (set): A set of numbers.

        Returns:
            set: A set of intervals between the successive numbers.
        '''
        S = sorted(S)
        return set(np.diff(S))

    @staticmethod
    def interval_vector(set1: set, modulus: int = 12) -> np.ndarray:        
        '''
        Compute the interval vector of a set of pitches.

        The interval vector represents the number of occurrences of each interval between pitches in a set.
        Intervals larger than half the modulus are inverted to their complements.

        Args:
            set1 (set): A set of integers.
            modulus (int): The modulus to use for interval calculations, conventionally 12.

        Returns:
            np.ndarray: An array representing the interval vector.
        '''
        pitches = sorted(set1)
        intervals = np.zeros(modulus // 2, dtype=int)

        for pitch1, pitch2 in combinations(pitches, 2):
            interval = abs(pitch2 - pitch1)
            interval = min(interval, modulus - interval)
            intervals[interval - 1] += 1

        return intervals


# ------------------------------------------------------------------------------
# Sieves
# ------

class Sieve:
    def __init__(self, modulus: int = 1, residue: int = 0, N: int = 255):
        self.__S = set(np.arange(residue, N + 1, modulus))
        self.__N = N
        self.__modulus = modulus
        self._residue = residue
    
    @property
    def S(self):
        return self.__S
    
    @property
    def N(self):
        return self.__N
    
    @property
    def period(self):
        return self.__modulus
    
    @property
    def r(self):
        return self._residue

    @property
    def congr(self):
        return Operations.congruent(self.__S, self.__modulus, self._residue)
    
    @property
    def compl(self):
        return Operations.complement(self.__S, self.__N)
    
    @N.setter
    def N(self, N: int):
        self.__N = N
        self.__S = set(np.arange(self._residue, N + 1, self.__modulus))
    
    def __str__(self) -> str:
        if len(self.__S) > 10:
            sieve = f'{list(self.__S)[:5]} ... {list(self.__S)[-1]}'
        else:
            sieve = list(self.__S)
        return (
            f'Period:  {self.__modulus}\n'
            f'Residue: {self._residue}\n'
            f'N:       {self.__N}\n'
            f'Sieve:   {sieve}\n'
        )

    def _repr__(self) -> str:        
        return self.__str__()
    

# ------------------------------------------------------------------------------
#  Generated Collection
# --------------

class GenCol:
    """
    Generated Collection - A multiplicative collection formed by repeatedly
    applying a generator within a periodic space.
    
    The collection is created by starting with an initial value (default 1)
    and repeatedly multiplying by the generator, reducing modulo the period.
    
    This is the multiplicative analog to the Sieve class.
    """
    def __init__(self, generator: Union[str,int,float,Fraction], period: Union[str,int,float,Fraction] = 2, iterations: int = 12):
        """
        Initialize a Generated Collection.
        
        Args:
            generator: The generator value (as str, int, float, or Fraction)
            period: The period of equivalence (as str, int, float, or Fraction)
            iterations: Number of times to apply the generator
        """
        self._generator = Fraction(generator)
        self._period = Fraction(period)
        self._iterations = iterations
        self._generate()
        
    @property
    def generator(self) -> Fraction:
        return self._generator
    
    @property
    def period(self) -> Fraction:
        return self._period
    
    @property
    def iterations(self) -> int:
        return self._iterations
    
    @property
    def collection(self) -> List[Fraction]:
        return self._collection.copy()
    
    @property
    def normalized_collection(self) -> List[Fraction]:
        if not hasattr(self, '_normalized_collection'):
            normalized = []
            for value in self._collection:
                current = value
                while current >= self._period:
                    current /= self._period
                normalized.append(current)
            self._normalized_collection = normalized
        return self._normalized_collection.copy()
    
    @property
    def steps(self) -> Set[Fraction]:
        if not hasattr(self, '_steps'):
            values = sorted(self.normalized_collection)
            steps = set()
            
            for i in range(len(values)):
                if i < len(values) - 1:
                    steps.add(values[i+1] / values[i])
                else:
                    steps.add(self._period * values[0] / values[i])
            
            self._steps = steps
        return self._steps.copy()
    
    def _generate(self):
        self._collection = [self._generator ** i for i in range(self._iterations + 1)]
        
        if hasattr(self, '_normalized_collection'):
            delattr(self, '_normalized_collection')
        if hasattr(self, '_steps'):
            delattr(self, '_steps')
    
    def __str__(self):
        result = (
            f"Generated Collection\n"
            f"Generator: {self._generator}\n"
            f"Period: {self._period}\n"
            f"Iterations: {self._iterations}\n"
        )
        
        if len(self._collection) <= 20:
            result += f"Values: {', '.join([str(s) for s in self.normalized_collection])}\n"
            result += f"Steps: {', '.join([str(s) for s in sorted(self.steps)])}\n"
        
        return result
    
    def __repr__(self):
        return self.__str__()


# --------------------------------------------------------------------------------
# Combination Sets (CS)
# ---------------------

class CombinationSet:
    '''
    A class representing a combination set where elements from a set of factors
    are combined according to a specified rank parameter.
    
    This class generates all combinations of size r from a set of factors.
    '''
    def __init__(self, factors:tuple = ('A', 'B', 'C', 'D'), r:int = 2):
        self._factors = tuple(sorted(factors))
        self._r = r
        self._combos = set(combinations(self._factors, self._r))
        self._factor_aliases = {f: sp.Symbol(chr(65 + i)) for i, f in enumerate(self._factors)}
        self._graph = self._generate_graph()

    @property
    def factors(self):
        return self._factors
    
    @property
    def rank(self):
        return self._r
    
    @property
    def combos(self):
        return self._combos
    
    @property
    def graph(self):
        return self._graph
    
    @property
    def factor_to_alias(self):
        return self._factor_aliases
    
    @property
    def alias_to_factor(self):
        return {v: k for k, v in self._factor_aliases.items()}
    
    def _generate_graph(self):
        G = nx.DiGraph()
        for i, combo in enumerate(self._combos):
            G.add_node(i, combo=combo)
        return G
    
    def __str__(self):    
        return (        
            f'Rank:    {self._r}\n'
            f'Factors: {self._factors}\n'
            f'Combos:  {self._combos}\n'
        )
    
    def _repr__(self) -> str:
        return self.__str__()


# ------------------------------------------------------------------------------
#  Partition Set
# --------------

class PartitionSet:
    def __init__(self, n: int, k: int):
        self._n = n
        self._k = k
        self._data = self._generate_partitions()
    
    def _generate_partitions(self) -> pd.DataFrame:
        '''
        Generate all possible partitions of n into k parts using backtracking,
        computing metadata for each partition as it's generated.
        
        Returns:
            pd.DataFrame: DataFrame containing partitions and their metadata.
        '''
        def backtrack(remaining: int, k: int, start: int, current: tuple) -> list:
            if k == 0:
                if remaining == 0:
                    return [{
                        'partition': current,
                        'unique_count': len(set(current)),
                        'span': max(current) - min(current),
                        'variance': np.var(current)
                    }]
                return []
            
            results = []
            for x in range(start, 0, -1):
                if x <= remaining:
                    results.extend(backtrack(remaining - x, k - 1, x, current + (x,)))
            return results
                    
        return pd.DataFrame(backtrack(self._n, self._k, self._n, ()))
    
    @property
    def data(self):
        return self._data
    
    @property
    def partitions(self):
        return tuple(self._data['partition'])
    
    @property
    def mean(self) -> float:
        '''The mean value of every partition in this set.'''
        return self._n / self._k

    def __str__(self) -> str:
        display_df = self._data.copy()
        display_df['variance'] = display_df['variance'].round(4)
        
        df_str = str(display_df)
        width = max(len(line) for line in df_str.split('\n'))
        border = '-' * width
        
        header = (
            f"{border}\n"
            f"PS(n={self._n}, k={self._k})\n"
            f"Mean: ~{round(self.mean, 4)}\n"
            f"{border}\n"
        )
        return header + df_str + f"\n{border}\n"
    
    def __repr__(self) -> str:
        return self.__str__()




# ------------------------------------------------------------------------------------
# Klotho/klotho/aikous/dynamics.py
# ------------------------------------------------------------------------------------
'''
--------------------------------------------------------------------------------------
Classes and functions for working with musical dynamics.
--------------------------------------------------------------------------------------
'''

import numpy as np
from numpy.polynomial import Polynomial
from scipy import interpolate

__all__ = [
    'DynamicRange',
    'ampdb',
    'dbamp',
    # 'amp_freq_scale',
    'freq_amp_scale',
]

DYNAMIC_MARKINGS = ('ppp', 'pp', 'p', 'mp', 'mf', 'f', 'ff', 'fff')

class Dynamic:
    def __init__(self, marking, db_value):
        self._marking = marking
        self._db_value = db_value
    
    @property
    def marking(self):
        return self._marking
    
    @property
    def db(self):
        return self._db_value
        
    @property
    def amp(self):
        return dbamp(self._db_value)
    
    def __float__(self):
        return float(self._db_value)
    
    def __repr__(self):
        return f"Dynamic(marking='{self._marking}', db={self._db_value:.2f}, amp={self.amp:.4f})"


class DynamicRange:
  '''
  Musical dynamics mapped to decibels.

  Note: the decibel level for the loudest 
  dynamic (ffff) is 0 dB as this translates 
  to an amplitude of 1.0.

  ----------------|---------|----------------
  Name            | Letters	| Level
  ----------------|---------|----------------
  fortississimo	  | fff	    | very very loud  
  fortissimo	    | ff	    | very loud
  forte	          | f	      | loud
  mezzo-forte	    | mf	    | moderately loud
  mezzo-piano	    | mp	    | moderately quiet
  piano	          | p	      | quiet
  pianissimo	    | pp	    | very quiet
  pianississimo	  | ppp	    | very very quiet
  ----------------|---------|----------------

  see https://en.wikipedia.org/wiki/Dynamics_(music)#
  '''
  def __init__(self, min_dynamic=-60, max_dynamic=-3, curve=0, dynamics=DYNAMIC_MARKINGS):
    self._min_db = min_dynamic
    self._max_db = max_dynamic
    self._curve = curve
    self._dynamics = dynamics
    self._range = self._calculate_range()
    
  @property
  def min_db(self):
    return self._min_db
  
  @property
  def max_db(self):
    return self._max_db
  
  @property
  def curve(self):
    return self._curve
  
  @property
  def ranges(self):
    return self._range

  def _calculate_range(self):
    min_db = float(self._min_db)
    max_db = float(self._max_db)
    num_dynamics = len(self._dynamics)
    
    result = {}
    for i, dyn in enumerate(self._dynamics):
        normalized_pos = i / (num_dynamics - 1)
        
        if self._curve == 0:
            curved_pos = normalized_pos
        elif self._curve > 0:
            curved_pos = normalized_pos ** (1 + self._curve)
        else:
            curved_pos = 1 - ((1 - normalized_pos) ** (1 - self._curve))
            
        db_value = min_db + curved_pos * (max_db - min_db)
        result[dyn] = Dynamic(dyn, db_value)
        
    return result

  def __getitem__(self, dynamic):
    return self._range[dynamic]

def ampdb(amp: float) -> float:
  '''
  Convert amplitude to decibels (dB).

  Args:
  amp (float): The amplitude to convert.

  Returns:
  float: The amplitude in decibels.
  '''
  return 20 * np.log10(amp)

def dbamp(db: float) -> float:
  '''
  Convert decibels (dB) to amplitude.

  Args:
  db (float): The decibels to convert.

  Returns:
  float: The amplitude.
  '''
  return 10 ** (db / 20)

# def amp_freq_scale(freq: float,
#     freqs: list = [20,  100, 250, 500, 1000, 2000, 3000, 4000, 6000, 10000, 20000],
#     amps: list  = [0.3, 0.5, 0.7, 0.8, 0.5,  0.6,  0.5,  0.6,  0.7,  0.5,   0.3],
#     deg: int = 4) -> float:
#   frequencies_sample = np.array(freqs, dtype=float)
#   loudness_sample    = np.array(amps, dtype=float)
#   p = Polynomial.fit(frequencies_sample, loudness_sample, deg=deg)
#   return p(freq)

def freq_amp_scale(freq: float, db_level: float, min_db: float = -60) -> float:
    """
    Scale amplitude based on frequency and loudness according to psychoacoustic principles.
    
    Args:
        freq (float): The frequency in Hz
        db_level (float): The input level in dB
        min_db (float): The minimum dB level in the dynamic range (default -60)
        
    Returns:
        float: The perceptually scaled amplitude (linear scale)
    """
    range_db = abs(min_db)
    phon_level = 40 + ((db_level - min_db) / range_db) * 60
    
    frequencies = np.array([20, 50, 100, 200, 500, 1000, 2000, 5000, 10000, 20000], dtype=float)
    
    if phon_level <= 40:
        scaling_curve = np.array([0.2, 0.3, 0.5, 0.7, 0.9, 1.0, 1.0, 0.9, 0.7, 0.4], dtype=float)
    elif phon_level <= 70:
        scaling_curve = np.array([0.3, 0.45, 0.6, 0.8, 0.95, 1.0, 1.0, 0.95, 0.8, 0.5], dtype=float)
    else:
        scaling_curve = np.array([0.5, 0.6, 0.7, 0.85, 0.95, 1.0, 1.0, 0.95, 0.85, 0.6], dtype=float)
    
    spline = interpolate.CubicSpline(frequencies, scaling_curve, extrapolate=True)
    scaling_factor = max(0.01, float(spline(freq)))
    
    raw_amp = dbamp(db_level)
    return raw_amp * scaling_factor



from ...topos.graphs.trees import Tree
import pandas as pd

class ParameterTree(Tree):
    def __init__(self, root, children:tuple):
        super().__init__(root, children)
        self._meta['pfields'] = pd.Series([set()], index=[''])
    
    def __getitem__(self, node):
        return ParameterNode(self, node)
    
    @property
    def pfields(self):
        return sorted(self._meta.loc['', 'pfields'])
    
    def set(self, node, **kwargs):
        self._meta.loc['', 'pfields'].update(kwargs.keys())
        
        for key, value in kwargs.items():
            self.graph.nodes[node][key] = value
        
        for descendant in self.descendants(node):
            for key, value in kwargs.items():
                self.graph.nodes[descendant][key] = value
        
    def get(self, node, key):
        if key in self.graph.nodes[node]:
            return self.graph.nodes[node][key]
        return None
    
    def clear(self, node=None):
        if node is None:
            for n in self.graph.nodes:
                self.graph.nodes[n].clear()
        else:
            self.graph.nodes[node].clear()
            for descendant in self.descendants(node):
                self.graph.nodes[descendant].clear()
            
    def items(self, node):
        return dict(self.graph.nodes[node])

class ParameterNode:
    def __init__(self, tree, node):
        self._tree = tree
        self._node = node
        
    def __getitem__(self, key):
        if isinstance(key, str):
            return self._tree.get(self._node, key)
        raise TypeError("Key must be a string")
    
    def __setitem__(self, key, value):
        self._tree.set(self._node, **{key: value})
        
    def clear(self):
        self._tree.clear(self._node)
        
    def items(self):
        return self._tree.items(self._node)
    
    def __dict__(self):
        return self._tree.items(self._node)
        
    def __str__(self):
        return str(self._tree.items(self._node))
    
    def __repr__(self):
        return repr(self._tree.items(self._node))


# import cairo
# from klotho.chronos.temporal_units.temporal import TemporalUnit

# def save_temporal_unit_image(temporal_unit: TemporalUnit, filename: str, width: int = 500, height: int = 25) -> None:
#     """
#     Saves a temporal unit visualization as a PDF image.
    
#     Args:
#         temporal_unit: The TemporalUnit to visualize
#         filename: Path where to save the image (should end in .pdf)
#         width: Width in points (1/72 inch)
#         height: Height in points
#     """
#     # Colors (RGB values from 0 to 1)
#     BLACK = (0, 0, 0)
#     GRAY = (0.5, 0.5, 0.5)
#     WHITE = (1, 1, 1)
#     BORDER = 1
    
#     # Create PDF surface
#     surface = cairo.PDFSurface(filename, width, height)
#     ctx = cairo.Context(surface)
    
#     # Calculate total duration and positions
#     total_duration = sum(abs(d) for d in temporal_unit.durations)
#     current_x = 0
    
#     for duration in temporal_unit.durations:
#         # Calculate width for this segment
#         block_width = (abs(duration) / total_duration) * width
#         if block_width == 0:
#             continue
        
#         # Set fill color and draw rectangle
#         if duration < 0:
#             ctx.set_source_rgb(*GRAY)
#         else:
#             ctx.set_source_rgb(*BLACK)
        
#         ctx.rectangle(current_x, 0, block_width, height)
#         ctx.fill()
        
#         # Draw white border
#         if current_x > 0:  # Draw left border if not first segment
#             ctx.set_source_rgb(*WHITE)
#             ctx.set_line_width(BORDER)
#             ctx.move_to(current_x, 0)
#             ctx.line_to(current_x, height)
#             ctx.stroke()
        
#         current_x += block_width
    
#     # Clean up
#     surface.finish()



from abc import ABC, abstractmethod
from klotho.aikous.expression.dynamics import Dynamic, DynamicRange
from klotho.tonos.pitch import Pitch
from klotho.utils.data_structures.dictionaries import SafeDict
from typing import List, Dict, TypeVar, Union

class Instrument(ABC):
    
    def __init__(self, name, freq_range=None, dynamic_range=None, pfields=None):
        """
        Initialize an Instrument.
        
        Args:
            name (str): The name of the instrument
            freq_range (tuple): A tuple of (min, max) frequency values or Pitch instances
            dynamic_range: A DynamicRange instance or a tuple of (min, max) dB values
            pfields (dict or SafeDict): Parameter fields with default values
        """
        self._name = name
        
        if freq_range is None:
            freq_range = (20.0, 20000.0)
        self._freq_range = self._process_freq_range(freq_range)
        
        if dynamic_range is None:
            dynamic_range = (-60, 0)
        self._dynamic_range = self._process_dynamic_range(dynamic_range)
        
        if pfields is None:
            pfields = {}
        self._pfields = pfields if isinstance(pfields, SafeDict) else SafeDict(pfields)
    
    def _process_freq_range(self, freq_range):
        min_freq, max_freq = freq_range
        
        if not isinstance(min_freq, Pitch):
            min_freq = Pitch.from_freq(float(min_freq))
        
        if not isinstance(max_freq, Pitch):
            max_freq = Pitch.from_freq(float(max_freq))
            
        return (min_freq, max_freq)
    
    def _process_dynamic_range(self, dynamic_range):
        if isinstance(dynamic_range, DynamicRange):
            return dynamic_range
        
        min_dyn, max_dyn = dynamic_range
        
        if isinstance(min_dyn, Dynamic):
            min_dyn = min_dyn.db
            
        if isinstance(max_dyn, Dynamic):
            max_dyn = max_dyn.db
            
        return DynamicRange(min_dynamic=min_dyn, max_dynamic=max_dyn)
    
    @property
    def name(self):
        return self._name
    
    @property
    def freq_range(self):
        return self._freq_range
    
    @property
    def dynamic_range(self):
        return self._dynamic_range
    
    @property
    def pfields(self):
        return self._pfields
    
    def __repr__(self):
        # return f"Instrument(name='{self._name}', freq_range={self._freq_range}, dynamic_range={self._dynamic_range}, pfields={dict(self._pfields)})"
        return f"Instrument(name='{self._name}', pfields={dict(self._pfields)})"



from typing import List, Callable, TypeVar, Any
import pandas as pd
import numpy as np
import networkx as nx
from networkx.algorithms.approximation import greedy_tsp

T = TypeVar('T')

def normalize_sum(values: List[float]) -> List[float]:
    """
    Normalize a list of positive values so they sum to 1 while maintaining their proportions.
    
    Args:
        values: List of positive numerical values
        
    Returns:
        List of normalized values that sum to 1
    """
    total = sum(values)
    if total <= 0:
        raise ValueError("Sum of values must be positive")
    return [v / total for v in values]

def cost_matrix(items: List[T], cost_function: Callable[[T, T], float], **kwargs: Any) -> pd.DataFrame:
    cost_ufunc = np.frompyfunc(lambda a, b: cost_function(a, b, **kwargs), 2, 1)
    arr = cost_ufunc(np.array(items)[:, None], np.array(items)[None, :]).astype(float)
    df = pd.DataFrame(arr, columns=[str(item) for item in items])
    df.insert(0, "", [str(item) for item in items])
    df = df.set_index("").rename_axis(None)
    return df

def cost_matrix_graph(cost_df: pd.DataFrame) -> nx.Graph:
    """
    Convert a cost matrix DataFrame to an undirected networkx graph.
    Edge weights represent the costs between nodes.
    Self-loops are excluded.
    
    Args:
        cost_df: DataFrame containing the cost matrix
        
    Returns:
        networkx Graph with weighted edges
    """
    G = nx.Graph()
    
    for i, value in enumerate(cost_df.columns):
        G.add_node(i, value=value)
    
    for i in range(len(cost_df.columns)):
        for j in range(i + 1, len(cost_df.columns)):
            cost = cost_df.iloc[i, j]
            if cost > 0:
                G.add_edge(i, j, weight=cost)
    
    return G

def minimum_cost_path(G: nx.Graph, start: int, end: int | None = None) -> List[int]:
    if start not in G.nodes():
        raise ValueError(f"Start node {start} not in graph")
    
    if end is not None:
        if end not in G.nodes():
            raise ValueError(f"End node {end} not in graph")
        
        nodes_to_visit = set(G.nodes()) - {start, end}
        path = [start]
        current = start
        
        while nodes_to_visit:
            next_node = min(nodes_to_visit, key=lambda x: G[current][x]['weight'])
            path.append(next_node)
            nodes_to_visit.remove(next_node)
            current = next_node
            
        path.append(end)
        return path
    
    return greedy_tsp(G, source=start)



import random

def rand_encode(keys, values, allow_repeats=False):
    '''
    Assigns random keys from the values list to the keys list. Allows for an option to repeat values.

    Args:
    - keys (list): A list of string symbols to be used as dictionary keys.
    - values (list): A list of string symbols to be assigned randomly to the keys.
    - allow_repeats (bool): If True, values will be repeated by re-shuffling once all are used.

    Returns:
    - dict: A dictionary with keys from the keys list and random values from the values list.
    '''
    if not keys or not values:
        return {}  # Return an empty dictionary if either list is empty

    assignments = {}
    values_pool = values.copy()
    random.shuffle(values_pool)

    for i, key in enumerate(keys):
        if not values_pool:  # If values_pool is empty
            if not allow_repeats:
                break  # Stop assigning if repeats are not allowed
            values_pool = values.copy()  # Replenish the values pool
            random.shuffle(values_pool)

        # Assign a value to the key
        assignments[key] = values_pool.pop()

    return assignments

# def linear_encode(keys, values, allow_repeats=False):
#     '''
#     Assigns values from the values list to the keys list in a linear fashion. Allows for an option to repeat values.
    
#     Args:
#     - keys (list): A list of string symbols to be used as dictionary keys.
#     - values (list): A list of string symbols to be assigned linearly to the keys.
#     - allow_repeats (bool): If True, values will be repeated by re-shuffling once all are used.
    
#     Returns:
#     - dict: A dictionary with keys from the keys list and values from the values list.
#     '''
#     if not keys or not values:
#         return {}  # Return an empty dictionary if either list is empty




"""
Klotho: A graph-oriented Python package for computational composition.

This package provides tools for working with musical structures across multiple domains:
- Topos: Abstract structures and relationships
- Chronos: Time and rhythm
- Tonos: Pitch and harmony
- Aikous: Expression and parameters
- Skora: Visualization and notation
"""
from . import topos
from . import chronos
from . import tonos
from . import aikous
from . import skora
from . import utils

from .topos.collections import patterns, sequences, sets, Pattern, CombinationSet, PartitionSet
from .topos.graphs import trees, networks, fields, Tree, Network, Field, Graph

from .chronos import RhythmPair, RhythmTree, TemporalUnit, TemporalUnitSequence, TemporalBlock

from .tonos import Pitch, Scale, Chord, AddressedScale, AddressedChord

from .skora.visualization.plots import plot_tree, plot_rt, plot_ratios, plot_curve, plot_cps, plot_graph

from .types import frequency, cent, midicent, midi, amplitude, decibel, onset, duration

from .utils.playback.player import play, pause, stop, sync
from .utils.playback.midi_export import midi as export_midi

__all__ = [
    'topos', 'chronos', 'tonos', 'aikous', 'skora', 'utils',
    'Pitch', 'Scale', 'Chord', 
    'AddressedScale', 'AddressedChord',
    'frequency', 'cent', 'midicent', 'midi',
    'amplitude', 'decibel', 'onset', 'duration',
    'play', 'pause', 'stop', 'sync', 'export_midi'
]

__version__ = '2.2.5'



import numpy as np
from typing import Union, List, Tuple, TypeVar, Any

from .tonos.utils.frequency_conversion import freq_to_midicents, midicents_to_freq
from .aikous.expression.dynamics import ampdb, dbamp

T = TypeVar('T', bound='float')
Collection = Union[List[float], Tuple[float, ...], np.ndarray]

class frequency(float):
    def __new__(cls, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            return type(value)(cls(x) for x in value)
        return super().__new__(cls, value)
    
    def to_midicent(self):
        return midicent(freq_to_midicents(float(self)))
    
    def to_midi(self):
        return midi(self.to_midicent() / 100)

class midi(float):
    def __new__(cls, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            return type(value)(cls(x) for x in value)
        return super().__new__(cls, value)
    
    def to_frequency(self):
        return frequency(midicents_to_freq(float(self) * 100))
    
    def to_midicent(self):
        return midicent(float(self) * 100)

class midicent(float):
    def __new__(cls, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            return type(value)(cls(x) for x in value)
        return super().__new__(cls, value)
    
    def to_midi(self):
        return midi(float(self) / 100)
    
    def to_frequency(self):
        return frequency(midicents_to_freq(float(self)))

class cent(float):
    def __new__(cls, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            return type(value)(cls(x) for x in value)
        return super().__new__(cls, value)
    
    def to_frequency_ratio(self):
        return 2.0 ** (float(self) / 1200.0)

class amplitude(float):
    def __new__(cls, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            return type(value)(cls(x) for x in value)
        return super().__new__(cls, value)
    
    def to_decibel(self):
        return decibel(ampdb(float(self)))

class decibel(float):
    def __new__(cls, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            return type(value)(cls(x) for x in value)
        return super().__new__(cls, value)
    
    def to_amplitude(self):
        return amplitude(dbamp(float(self)))

class onset(float):
    def __new__(cls, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            return type(value)(cls(x) for x in value)
        return super().__new__(cls, value)

class duration(float):
    def __new__(cls, value):
        if isinstance(value, (list, tuple, np.ndarray)):
            return type(value)(cls(x) for x in value)
        return super().__new__(cls, value) 


"""
Skora: A specialized module for visualization, notation, and representation of music.

From the Greek "σκορ" (skor) related to "score" or "notation," this module
provides tools for visualizing and notating musical structures.
"""
from .skora import *

from . import animation
from . import notation
from . import visualization

from .visualization import plots
from .visualization.plots import *

__all__ = []



from klotho.topos.graphs.fields import Field
import numpy as np
import matplotlib.pyplot as plt
from typing import List, Tuple

def plot_field_heatmap(field: Field, title: str = '', path: List[Tuple[Tuple[float, float], float]] = None, save_path: str = None):
    """
    Plot the field as a heatmap with an optional navigation path overlaid.
    
    :param field: The Field object to plot
    :param title: Title for the plot
    :param path: Optional list of (point, value) tuples representing the navigation path
    """
    points = np.array(list(field.nodes.keys()))
    values = np.array(list(field.nodes.values()))

    grid_shape = (field.resolution,) * field.dimensionality

    Z = values.reshape(grid_shape)

    plt.figure(figsize=(12, 10), facecolor='black')
    plt.gcf().set_facecolor('black')
    plt.gca().set_facecolor('black')

    extent = [-1, 1, -1, 1] 
    plt.imshow(Z, extent=extent, origin='lower', cmap='plasma')
    
    cbar = plt.colorbar(label='Field Value')
    cbar.ax.yaxis.label.set_color('white')
    cbar.ax.tick_params(color='white')
    
    plt.title(title, color='white', fontsize=16)
    plt.xlabel('X', color='white')
    plt.ylabel('Y', color='white')
    plt.tick_params(colors='white')

    if path:
        path_points, path_values = zip(*path)
        path_x, path_y = zip(*path_points)
        plt.plot(path_x, path_y, color='white', linewidth=2, alpha=0.7)
        plt.scatter(path_x[0], path_y[0], color='lime', s=100, label='Start')
        plt.scatter(path_x[-1], path_y[-1], color='red', s=100, label='End')
        plt.legend(loc='upper right', facecolor='black', edgecolor='white', labelcolor='white')

    if save_path:
        plt.savefig(save_path)

    plt.tight_layout()
    plt.show()
    
def plot_path_color_bar(path, title=''):
    """
    Plot the path values as a horizontal color bar.
    
    :param path: List of (point, value) tuples representing the navigation path
    :param title: Title for the plot
    """
    path_values = [value for _, value in path]
    values = np.array(path_values).reshape(1, -1)
    
    plt.figure(figsize=(12, 4), facecolor='black')
    plt.gcf().set_facecolor('black')
    plt.gca().set_facecolor('black')
    
    img = plt.imshow(values, aspect='auto', cmap='plasma', extent=[0, len(path_values), 0, 1])
    
    plt.title(f"{title}", color='white', fontsize=16)
    plt.xlabel('Navigation Steps', color='white')
    plt.yticks([])  # Remove y-axis ticks
    plt.tick_params(colors='white')
    
    # cbar = plt.colorbar(img, orientation='horizontal', pad=0.2)
    # cbar.ax.xaxis.label.set_color('white')
    # cbar.ax.tick_params(color='white')
    # cbar.set_label('Field Value', color='white')
    
    plt.tight_layout()
    plt.show()


from klotho.topos.graphs.fields import Field
import random

# def find_navigation_path(field: Field, steps: int = 2000, seed: int = 42):
#     """
#     Generate a navigation path through the field.
    
#     :param field: The Field object to navigate
#     :param steps: Number of steps to take in the navigation
#     :return: List of (point, value) tuples representing the path
#     """
#     random.seed(seed)
#     start_point = random.choice(list(field.nodes.keys()))
#     path = [(start_point, field[start_point])]
#     visited = set([start_point])
    
#     for _ in range(steps - 1):
#         current_point = path[-1][0]
#         neighbors = field.get_neighbors(current_point)
#         unvisited_neighbors = [p for p in neighbors if p not in visited]
        
#         if unvisited_neighbors:
#             next_point = max(unvisited_neighbors, key=neighbors.get)
#             # next_point = random.choice(unvisited_neighbors)
#         elif neighbors:
#             next_point = random.choice(list(neighbors.keys()))
#         else:
#             break
        
#         path.append((next_point, field[next_point]))
#         visited.add(next_point)
    
#     return path

import numpy as np
from typing import List, Tuple

def find_navigation_path(field: Field, steps: int = 2000, frequency: float = 0.05) -> List[Tuple]:
    dimensions = field.dimensionality
    resolution = field.resolution
    path = []

    grid = np.meshgrid(*[np.arange(resolution) for _ in range(dimensions)])
    all_indices = np.stack([g.flatten() for g in grid], axis=-1)

    for t in range(steps):
        angle = frequency * t
        
        oscillating_point = np.zeros(dimensions)
        oscillating_point[0] = 0.5 * (1 + np.sin(angle) * np.cos(0.3 * angle)) * (resolution - 1)
        oscillating_point[1] = 0.5 * (1 + np.cos(1.5 * angle) * np.sin(0.4 * angle)) * (resolution - 1)
        
        for i in range(2, dimensions):
            oscillating_point[i] = 0.5 * (1 + 0.1 * np.sin(angle / (i + 1))) * (resolution - 1)

        distances = np.linalg.norm(all_indices - oscillating_point, axis=1)
        nearest_index = np.argmin(distances)
        nearest_point = tuple(all_indices[nearest_index])

        actual_point = tuple(field.nodes.keys())[nearest_index]
        path.append((actual_point, field[actual_point]))

    return path



"""
Aikous: A specialized module for working with expression and parameters in music.

From the Greek "ακούω" (akoúō) meaning "to hear" or "to listen," this module
deals with the expressive aspects of music that affect how we perceive sound.
"""

from . import expression
from . import parameters
from . import instruments

from .expression import Dynamic, DynamicRange
from .parameters import ParameterTree
from .instruments import Instrument

from .expression import dbamp, ampdb, freq_amp_scale
from .expression import line, arch, map_curve

__all__ = [
    # Modules
    'expression',
    'parameters',
    'instruments',
    
    # Classes
    'DynamicRange',
    'Dynamic',
    'ParameterTree',
    'Instrument',
    
    # Expression functions
    'dbamp', 
    'ampdb', 
    'freq_amp_scale',
    'line', 
    'arch', 
    'map_curve',
]



from fractions import Fraction
from sympy import factorint, prime as sympy_prime
from typing import Union, Dict
from functools import lru_cache

def to_factors(value: Union[int, Fraction, str]) -> Dict[int, int]:
    match value:
        case int() as i:
            ratio = Fraction(i, 1)
        case Fraction() as f:
            ratio = f
        case str() as s:
            ratio = Fraction(s)
        case _:
            raise TypeError("Unsupported type")
    num_factors = factorint(ratio.numerator)
    den_factors = factorint(ratio.denominator)
    for p, e in den_factors.items():
        num_factors[p] = num_factors.get(p, 0) - e
    return num_factors

def from_factors(factors: Dict[int, int]) -> Fraction:
    numerator = 1
    denominator = 1
    for prime, exp in factors.items():
        if exp > 0:
            numerator *= prime ** exp
        elif exp < 0:
            denominator *= prime ** (-exp)
    return Fraction(numerator, denominator)

@lru_cache(maxsize=128)
def prime_lattice_vector(ratio: Union[int, Fraction, str]) -> list[int]:
    factors = to_factors(Fraction(ratio))
    max_prime = max(factors.keys(), default=2)
    
    nth = 1
    while sympy_prime(nth) != max_prime:
        nth = nth + 1
    
    primes = [sympy_prime(i) for i in range(1, nth+1)]
    return [factors.get(p, 0) for p in primes]



from typing import Callable, Tuple, List, Dict
import numpy as np

class Field:
    def __init__(self, dimensionality: int, resolution: int, function: Callable[[np.ndarray], np.ndarray]):
        '''
        Initialize a Field.

        :param dimensionality: Number of dimensions for the field
        :param resolution: Number of points along each dimension
        :param function: Function to evaluate at each point
        '''
        self.dimensionality = dimensionality
        self.resolution = resolution
        self.nodes = {}
        self.edges = {}

        axes = [np.linspace(-1, 1, resolution) for _ in range(dimensionality)]
        grid = np.meshgrid(*axes)
        points = np.stack([ax.flatten() for ax in grid], axis=-1)

        if function is not None:
            values = function(points)
        else:
            values = np.zeros(resolution**dimensionality)
        # values = function(points)

        self.nodes = {tuple(point): float(value) for point, value in zip(points, values)}

        self._add_edges()

    def _add_edges(self):
        '''Add edges to form a grid with diagonals.'''
        points = list(self.nodes.keys())
        for i, point in enumerate(points):
            neighbors = self._get_neighbor_indices(i)
            for neighbor_idx in neighbors:
                if 0 <= neighbor_idx < len(points):
                    neighbor = points[neighbor_idx]
                    self.add_edge(point, neighbor)

    def _get_neighbor_indices(self, index: int) -> List[int]:
        '''Get the indices of neighboring points in n-dimensional space.'''
        coords = np.unravel_index(index, [self.resolution] * self.dimensionality)
        offsets = np.array(list(np.ndindex((3,) * self.dimensionality))) - 1
        neighbor_coords = np.array(coords) + offsets
        valid_coords = np.all((0 <= neighbor_coords) & (neighbor_coords < self.resolution), axis=1)
        return np.ravel_multi_index(neighbor_coords[valid_coords].T, [self.resolution] * self.dimensionality)

    def add_edge(self, point1: Tuple, point2: Tuple):
        '''Add an edge between two points.'''
        if point1 not in self.edges:
            self.edges[point1] = set()
        if point2 not in self.edges:
            self.edges[point2] = set()
        self.edges[point1].add(point2)
        self.edges[point2].add(point1)

    def get_field_value(self, point: Tuple) -> float:
        '''Get the field value at a specific point.'''
        return self.nodes[point]

    def get_neighbors(self, point: Tuple) -> Dict[Tuple, float]:
        '''Get the neighbors of a specific point.'''
        return {neighbor: self.nodes[neighbor] for neighbor in self.edges.get(point, set())}

    def __getitem__(self, point: Tuple) -> float:
        '''Allow field[point] access to field values.'''
        return self.get_field_value(point)

    def __setitem__(self, point: Tuple, value: float):
        '''Allow field[point] = value setting of field values.'''
        self.nodes[point] = value

    @classmethod
    def interact(cls, field1: 'Field', field2: 'Field', interaction_function: Callable[[float, float], float]) -> 'Field':
        '''
        Create a new field by interacting two existing fields.

        :param field1: First Field instance
        :param field2: Second Field instance
        :param interaction_function: Function that takes two field values and returns a new value
        :return: A new Field instance resulting from the interaction
        '''
        if field1.dimensionality != field2.dimensionality or field1.resolution != field2.resolution:
            raise ValueError("The two fields must have the same dimensionality and resolution")

        if set(field1.nodes.keys()) != set(field2.nodes.keys()):
            raise ValueError("The two fields must have the same space points")

        new_field = cls(field1.dimensionality, field1.resolution, None)
        
        # new_field.nodes = {
        #     point: interaction_function(field1[point], field2[point])
        #     for point in field1.nodes.keys()
        # }
        new_field.nodes = {
            point: interaction_function(field1, field2, point)
            for point in field1.nodes.keys()
        }

        new_field.edges = field1.edges.copy()

        return new_field

    def __str__(self) -> str:
        '''String representation of the field.'''
        return f"Field(dimensionality={self.dimensionality}, resolution={self.resolution}, nodes={len(self.nodes)}, edges={sum(len(e) for e in self.edges.values())//2})"

    def __repr__(self) -> str:
        return self.__str__()
    
import pickle

def save_field(field: Field, filename: str):
    """
    Save a Field instance to a file using pickle.
    
    Args:
    field (Field): The Field instance to save.
    filename (str): The name of the file to save the field to.
    """
    with open(filename, 'wb') as f:
        pickle.dump(field, f)
    print(f"Field saved to {filename}")

def load_field(filename: str) -> Field:
    """
    Load a Field instance from a file using pickle.
    
    Args:
    filename (str): The name of the file to load the field from.
    
    Returns:
    Field: The loaded Field instance.
    """
    with open(filename, 'rb') as f:
        field = pickle.load(f)
    print(f"Field loaded from {filename}")
    return field


from ...collections.sets import CombinationSet as CS

class Network:
    def __init__(self, nodes=None, edges=None):
        self._nodes = {}
        self._edges = {}
        
        if nodes:
            for node, attributes in nodes.items():
                self.add_node(node, **attributes)
        
        if edges:
            for edge, attributes in edges.items():
                self.add_edge(edge[0], edge[1], **attributes)
    
    def add_node(self, node, **attributes):
        self._nodes[node] = attributes
    
    def add_edge(self, node1, node2, **attributes):
        if node1 not in self._nodes or node2 not in self._nodes:
            raise ValueError("Both nodes must be added to the network before adding an edge.")
        
        if node1 not in self._edges:
            self._edges[node1] = {}
        if node2 not in self._edges:
            self._edges[node2] = {}

        self._edges[node1][node2] = attributes
        self._edges[node2][node1] = attributes
    
    @property
    def nodes(self):
        return self._nodes
    
    @property
    def edges(self):
        return self._edges

    def node_attributes(self, node):
        return self._nodes.get(node, {})
    
    def edge_attributes(self, node1, node2):
        return self._edges.get(node1, {}).get(node2, {})


class ComboNet:
    def __init__(self, cps:CS):
        self._cps = cps
        self._graph = self._make_network()

    def _make_network(self):
        graph = {}
        for combo in self._cps.combos:
            graph[combo] = {}

        for combo1 in self._cps.combos:
            for combo2 in self._cps.combos:
                if combo1 != combo2:
                    common_factors = len(set(combo1) & set(combo2))
                    if common_factors > 0:
                        graph[combo1][combo2] = common_factors
                        graph[combo2][combo1] = common_factors

        return graph

    @property
    def cps(self):
        return self._cps
    
    @property
    def graph(self):
        return self._graph

    def edge_weight(self, node1, node2):
        return self._graph.get(node1, {}).get(node2, 0)


import random
from collections import defaultdict
from ..networks import ComboNet
from klotho.tonos.combination_product_sets import CombinationProductSet
from typing import Union

class ComboNetTraversal:
    def __init__(self, combnet: Union[ComboNet, CombinationProductSet]):
        self.combnet = combnet if isinstance(combnet, ComboNet) else ComboNet(combnet)
        self.visited_edges = defaultdict(int)

    def play(self, start_node, num_steps):
        current_node = start_node
        path = [current_node]

        for _ in range(num_steps):
            next_node = self.choose_next_node(current_node)
            if next_node is None:
                print("No more available moves. Game over.")
                break
            
            path.append(next_node)
            self.visited_edges[self.edge_key(current_node, next_node)] += 1
            current_node = next_node

        return path

    def choose_next_node(self, current_node):
        neighbors = self.combnet.graph[current_node]
        if not neighbors:
            return None

        candidates = defaultdict(lambda: defaultdict(list))
        for neighbor, weight in neighbors.items():
            visit_count = self.visited_edges[self.edge_key(current_node, neighbor)]
            candidates[weight][visit_count].append(neighbor)

        sorted_weights = sorted(candidates.keys(), reverse=True)

        for weight in sorted_weights:
            visit_counts = sorted(candidates[weight].keys())
            for visit_count in visit_counts:
                if candidates[weight][visit_count]:
                    return random.choice(candidates[weight][visit_count])

        return None

    def edge_key(self, node1, node2):
        return tuple(sorted([node1, node2]))


# import networkx as nx
# import hypernetx as hnx
# from math import prod
# from fractions import Fraction

# from klotho.tonos.tonos import equave_reduce

# def graph_cps(combos:tuple):
#     G = nx.Graph()    
#     # each combination is a node (vertex) in the graph
#     for combo in combos:
#         G.add_node(combo)
#     # edges are between nodes that share at least one common factor
#     for combo1 in combos:
#         for combo2 in combos:
#             if combo1 != combo2 and set(combo1).intersection(combo2):
#                 G.add_edge(combo1, combo2)
#     return G

# def find_cliques(G:nx.Graph, n:int):
#     cliques = nx.enumerate_all_cliques(G)
#     return tuple(tuple(clique) for clique in cliques if len(clique) == n)

# def combo_clique_to_ratios(clique:tuple, min_prod:int, equave:int=2, n_equaves:int=1):
#     return tuple(sorted(equave_reduce(interval  = Fraction(prod(combo), min_prod),
#                                       equave    = equave,
#                                       n_equaves = n_equaves) for combo in clique))

# def hyper_graph_cps(G:nx.Graph, n):
#     n = [n] if not isinstance(n, list) else n
#     cliques = nx.enumerate_all_cliques(G)
#     edges = {}
#     for i, clique in enumerate(cliques):
#         if n_ad := len(clique) in n:
#             edges[f'{n_ad}-ad_{i}'] = clique
#     H = hnx.Hypergraph(edges)
#     return H
